@@ -51,7 +51,6 @@
 (defvar exwm-layout--timer nil "Timer used to track echo area changes.")
 
 (defvar exwm-workspace--current)
-(defvar exwm-workspace--frame-y-offset)
 (declare-function exwm-input--release-keyboard "exwm-input.el")
 (declare-function exwm-input--grab-keyboard "exwm-input.el")
 (declare-function exwm-input-grab-keyboard "exwm-input.el")
@@ -111,29 +110,22 @@ See variable `exwm-layout-auto-iconify'."
 (defun exwm-layout--show (id &optional window)
   "Show window ID exactly fit in the Emacs window WINDOW."
   (exwm--log "Show #x%x in %s" id window)
-  (let* ((edges (window-inside-absolute-pixel-edges window))
+  (let* ((edges (exwm--window-inside-absolute-pixel-edges window))
          (x (pop edges))
          (y (pop edges))
          (width (- (pop edges) x))
-         (height (- (pop edges) y))
-         frame-x frame-y frame-width frame-height)
-    (when (< emacs-major-version 31)
-      (setq y (+ y (window-tab-line-height window))))
+         (height (- (pop edges) y)))
     (with-current-buffer (exwm--id->buffer id)
-      (when exwm--floating-frame
-        (setq frame-width (frame-pixel-width exwm--floating-frame)
-              frame-height (+ (frame-pixel-height exwm--floating-frame)
-                              ;; Use `frame-outer-height' in the future.
-                              exwm-workspace--frame-y-offset))
-        (when exwm--floating-frame-position
-          (setq frame-x (elt exwm--floating-frame-position 0)
-                frame-y (elt exwm--floating-frame-position 1)
+      (when (and exwm--floating-frame exwm--floating-frame-geometry)
+        (with-slots ((frame-x x) (frame-y y)
+                     (frame-width width) (frame-height height))
+            exwm--floating-frame-geometry
+          (setq exwm--floating-frame-geometry nil
                 x (+ x frame-x (- exwm-layout--floating-hidden-position))
                 y (+ y frame-y (- exwm-layout--floating-hidden-position)))
-          (setq exwm--floating-frame-position nil))
-        (exwm--set-geometry (frame-parameter exwm--floating-frame
-                                             'exwm-container)
-                            frame-x frame-y frame-width frame-height))
+          (exwm--set-geometry (frame-parameter exwm--floating-frame
+                                               'exwm-container)
+                              frame-x frame-y frame-width frame-height)))
       (when (exwm-layout--fullscreen-p)
         (with-slots ((x* x)
                      (y* y)
@@ -167,8 +159,7 @@ See variable `exwm-layout-auto-iconify'."
                (geometry (xcb:+request-unchecked+reply exwm--connection
                              (make-instance 'xcb:GetGeometry
                                             :drawable container))))
-          (setq exwm--floating-frame-position
-                (vector (slot-value geometry 'x) (slot-value geometry 'y)))
+          (setq exwm--floating-frame-geometry geometry)
           (exwm--set-geometry container exwm-layout--floating-hidden-position
                               exwm-layout--floating-hidden-position
                               1
@@ -280,26 +271,25 @@ selected by `other-buffer'."
 (defun exwm-layout--set-client-list-stacking ()
   "Set _NET_CLIENT_LIST_STACKING."
   (exwm--log)
-  (let (id clients-floating clients clients-iconic clients-other)
-    (dolist (pair exwm--id-buffer-alist)
-      (setq id (car pair))
-      (with-current-buffer (cdr pair)
-        (if (eq exwm--frame exwm-workspace--current)
-            (if exwm--floating-frame
-                ;; A floating X window on the current workspace.
-                (setq clients-floating (cons id clients-floating))
-              (if (get-buffer-window (cdr pair) exwm-workspace--current)
-                  ;; A normal tilling X window on the current workspace.
-                  (setq clients (cons id clients))
-                ;; An iconic tilling X window on the current workspace.
-                (setq clients-iconic (cons id clients-iconic))))
-          ;; X window on other workspaces.
-          (setq clients-other (cons id clients-other)))))
+  (let (clients-floating clients clients-iconic clients-other)
+    (pcase-dolist (`(,id . ,buffer) exwm--id-buffer-alist)
+      (with-current-buffer buffer
+        (cond
+         ;; X window on other workspaces.
+         ((not (eq exwm--frame exwm-workspace--current))
+          (push id clients-other))
+         ;; A floating X window on the current workspace.
+         (exwm--floating-frame (push id clients-floating))
+         ;; A normal tilling X window on the current workspace.
+         ((get-buffer-window buffer exwm-workspace--current)
+          (push id clients))
+         ;; An iconic tilling X window on the current workspace.
+         (t (push id clients-iconic)))))
     (xcb:+request exwm--connection
         (make-instance 'xcb:ewmh:set-_NET_CLIENT_LIST_STACKING
                        :window exwm--root
-                       :data (vconcat (append clients-other clients-iconic
-                                              clients clients-floating))))))
+                       :data (vconcat clients-other clients-iconic
+                                      clients clients-floating)))))
 
 (defun exwm-layout--refresh (&optional frame)
   "Refresh layout of FRAME.
@@ -456,8 +446,8 @@ windows."
    ;; Resize on floating layout
    (exwm--fixed-size)                   ;fixed size
    (horizontal
-    (let* ((width (frame-pixel-width))
-           (edges (window-inside-pixel-edges))
+    (let* ((width (frame-outer-width))
+           (edges (exwm--window-inside-pixel-edges))
            (inner-width (- (elt edges 2) (elt edges 0)))
            (margin (- width inner-width)))
       (if (> delta 0)
@@ -489,8 +479,8 @@ windows."
                            :width width))
         (xcb:flush exwm--connection))))
    (t
-    (let* ((height (+ (frame-pixel-height) exwm-workspace--frame-y-offset))
-           (edges (window-inside-pixel-edges))
+    (let* ((height (frame-outer-height))
+           (edges (exwm--window-inside-pixel-edges))
            (inner-height (- (elt edges 3) (elt edges 1)))
            (margin (- height inner-height)))
       (if (> delta 0)
@@ -546,49 +536,104 @@ See also `exwm-layout-enlarge-window'."
   (exwm--log "%s" delta)
   (exwm-layout-enlarge-window (- delta) t))
 
+(defun exwm-layout--window-bottom-offset (window)
+  "Compute the distance from the bottom of WINDOW to the bottom of its frame."
+  (- (elt (frame-edges (window-frame window) 'outer-edges) 3)
+     (elt (exwm--window-inside-pixel-edges ) 3)))
+
 (defun exwm-layout-hide-mode-line ()
-  "Hide mode-line."
+  "Hide the mode-line.
+See `exwm-layout-toggle-mode-line' for more details."
   (interactive)
   (exwm--log)
-  (when (and (derived-mode-p 'exwm-mode) mode-line-format)
-    (let (mode-line-height)
-      (when exwm--floating-frame
-        (setq mode-line-height (window-mode-line-height
-                                (frame-root-window exwm--floating-frame))))
-      (setq exwm--mode-line-format mode-line-format
-            mode-line-format nil)
-      (if (not exwm--floating-frame)
-          (exwm-layout--show exwm--id)
-        (set-frame-height exwm--floating-frame
-                          (- (frame-pixel-height exwm--floating-frame)
-                             mode-line-height)
-                          nil t)))))
+  (exwm-layout-toggle-mode-line -1))
 
 (defun exwm-layout-show-mode-line ()
-  "Show mode-line."
+  "Show the mode-line.
+See `exwm-layout-toggle-mode-line' for more details."
   (interactive)
   (exwm--log)
-  (when (and (derived-mode-p 'exwm-mode) (not mode-line-format))
-    (setq mode-line-format exwm--mode-line-format
-          exwm--mode-line-format nil)
-    (if (not exwm--floating-frame)
-        (exwm-layout--show exwm--id)
-      (set-frame-height exwm--floating-frame
-                        (+ (frame-pixel-height exwm--floating-frame)
-                           (window-mode-line-height (frame-root-window
-                                                     exwm--floating-frame)))
-                        nil t)
-      (call-interactively #'exwm-input-grab-keyboard))
-    (force-mode-line-update)))
+  (exwm-layout-toggle-mode-line 1))
 
-(defun exwm-layout-toggle-mode-line ()
-  "Toggle the display of mode-line."
-  (interactive)
+;; You can do this with by let-binding places with `gv-ref' and
+;; `gv-deref' instead of a macro, but the macro is cleaner.
+(defmacro exwm-layout--toggle-mode-line-1
+    (hide active-place saved-place none)
+  "Macro implementing the core logic behind mode-line toggling.
+
+If HIDE, the modeline is hidden. Otherwise, it is shown.
+
+ACTIVE-PLACE is the generalized variable where the active mode-line
+is stored.
+
+SAVED-PLACE is the generalized variable where the saved mode-line is
+stored when hidden.
+
+NONE is the symbol stored in ACTIVE-PLACE to hide the mode-line."
+  `(if ,hide
+       (ignore (cl-shiftf ,saved-place ,active-place ',none))
+     (setf ,active-place (or (cl-shiftf ,saved-place nil)
+                          mode-line-format
+                          (default-value 'mode-line-format)
+                          (error "No sane mode-line to show")))))
+
+(defsubst exwm-layout--window-specific-mode-line (window)
+  "Return non-nil if WINDOW has a WINDOW-specific mode-line."
+  (or (window-parameter window 'mode-line-format)
+      (window-parameter window 'exwm--saved-mode-line-format)))
+
+(defsubst exwm-layout--mode-line (&optional window)
+  "Return the effective mode-line for WINDOW or the current buffer.
+Return nil if the buffer/window has no mode-line."
+  (pcase (and window (window-parameter window 'mode-line-format))
+    ('none nil)
+    ('nil mode-line-format)
+    (wml wml)))
+
+(defun exwm-layout-toggle-mode-line (&optional arg)
+  "Toggle the display of mode-line.
+
+If ARG is a positive number, show the mode-line.
+If ARG is a negative number, hide the mode-line.
+Otherwise, toggle the mode-line.
+
+If the mode-line format is specific to the current window (e.g., an
+undecorated floating window), the mode-line is toggled in that window
+only. Otherwise, it's toggled globally."
+  (interactive (list (and current-prefix-arg
+                          (prefix-numeric-value current-prefix-arg))))
   (exwm--log)
-  (when (derived-mode-p 'exwm-mode)
-    (if mode-line-format
-        (exwm-layout-hide-mode-line)
-      (exwm-layout-show-mode-line))))
+  (let* ((target-window
+          (cond (exwm--floating-frame
+                 (frame-first-window exwm--floating-frame))
+                ((eq (window-buffer) (current-buffer))
+                 (selected-window))))
+         (is-visible
+          (not (null (exwm-layout--mode-line target-window)))))
+    (when (or (not (numberp arg)) (eq (< arg 0) is-visible))
+      (let ((old-bottom-offset
+             (and exwm--floating-frame
+                  target-window
+                  (exwm-layout--window-bottom-offset target-window))))
+        (if (and target-window
+                 (exwm-layout--window-specific-mode-line target-window))
+            (exwm-layout--toggle-mode-line-1
+             is-visible
+             (window-parameter target-window
+                               'mode-line-format)
+             (window-parameter target-window
+                               'exwm--saved-mode-line-format)
+             none)
+          (exwm-layout--toggle-mode-line-1
+           is-visible
+           mode-line-format
+           exwm--saved-mode-line-format
+           nil))
+        (when old-bottom-offset
+          (exwm-layout-enlarge-window
+           (- (exwm-layout--window-bottom-offset target-window)
+              old-bottom-offset))))
+      (force-mode-line-update))))
 
 (defun exwm-layout--init ()
   "Initialize layout module."
